/*
 * stmt.c - Statement construction and execution
 *
 * Written 2002,2003 by Werner Almesberger, Caltech Netlab FAST project
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "dwarfx.h"

#include "util.h"
#include "data.h"
#include "array.h"
#include "var.h"
#include "error.h"
#include "io.h"
#include "sprintf.h"
#include "stmt.h"


static const struct stmt *ip = NULL;


/* ----- Construction and destruction -------------------------------------- */


static struct stmt *stmts; /* list, for deallocation */


struct stmt *new_stmt(void (*fn)(const struct stmt *self))
{
    struct stmt *stmt;

    stmt = alloc_t(struct stmt);
    stmt->fn = fn;
    stmt->next = NULL;
    stmt->more = stmts;
    stmt->loc = current_location();
    stmts = stmt;
    return stmt;
}


static void destroy_without(struct without *wo)
{
    while (wo) {
	struct without *next = wo->next;

	destroy_expr(wo->expr);
	free(wo);
	wo = next;
    }
}


static void destroy_stmt(struct stmt *stmt)
{
    if (stmt->fn == stmt_expression) destroy_expr(stmt->u.expr.expr);
    else if (stmt->fn == stmt_print) {
	destroy_arg_list(stmt->u.print.args);
	destroy_without(stmt->u.print.without);
    }
    else if (stmt->fn == stmt_printf) {
	destroy_expr(stmt->u.printf.fmt);
	destroy_arg_list(stmt->u.printf.args);
    }
    else if (stmt->fn == stmt_if) destroy_expr(stmt->u.if_stmt.expr);
    else if (stmt->fn == stmt_break) /* do nothing */;
    else if (stmt->fn == stmt_exit) destroy_expr(stmt->u.exit.expr);
    else if (stmt->fn == stmt_die) destroy_expr(stmt->u.die.expr);
    else if (stmt->fn == stmt_while) destroy_expr(stmt->u.while_stmt.expr);
    else abort();
    free(stmt);
}


void stmt_cleanup(void)
{
    while (stmts) {
	struct stmt *next = stmts->more;

	destroy_stmt(stmts);
	stmts = next;
    }
}


/* ----- Fix statement pointers -------------------------------------------- */


static void end_sequence(struct stmt **stmt,struct stmt *next)
{
    int more;

    if (!*stmt) {
	*stmt = next;
	return;
    }
    more = 1;
    while (more) {
	if (!(*stmt)->next) {
	    (*stmt)->next = next;
	    more = 0;
	}
	if ((*stmt)->fn == stmt_if) {
	    end_sequence(&(*stmt)->u.if_stmt.true,(*stmt)->next);
	    end_sequence(&(*stmt)->u.if_stmt.false,(*stmt)->next);
	}
	else if ((*stmt)->fn == stmt_while)
	    end_sequence(&(*stmt)->u.while_stmt.body,*stmt);
	else if ((*stmt)->fn == stmt_break)
	    (*stmt)->u.break_stmt.loop = (*stmt)->u.break_stmt.loop->next;
	stmt = &(*stmt)->next;
    }
}


void fix_statements(struct stmt **stmt)
{
    end_sequence(stmt,NULL);
}


/* ----- Statement execution ----------------------------------------------- */


void stmt_expression(const struct stmt *self)
{
    put(evaluate_lvalue(self->u.expr.expr));
}


void stmt_print(const struct stmt *self)
{
    const struct without *wo;
    const struct arg_list *arg;

    for (wo = self->u.print.without; wo; wo = wo->next) {
	struct value *value;

	value = evaluate(wo->expr);
	if (!value) clerrorf("undefined value in without list");
	dwx_output_skip(value->value,dwx_sizeof(value->type));
	put(value);
    }
    for (arg = self->u.print.args; arg; arg = arg->next) {
	struct value *value;

	value = evaluate(arg->expr);
	if (!value) clerrorf("don't know how to print undefined value");
	if (value->type == &string_type)
	    printf("%s\n",(const char *) value->value);
	else dwx_output_data(stdout,value->type,value->value);
	put(value);
    }
    dwx_output_all();
}


void stmt_printf(const struct stmt *self)
{
    const char *s;

    s = common_sprintf(self->u.printf.fmt,self->u.printf.args);
    printf("%s\n",s);
    free((char *) s);
}


#if 0
void stmt_read(const struct stmt *self)
{
    struct value value;
    const char *s;
    int n;
    const void *p;

    value = evaluate(self->u.read.expr);
    if (value.type == &type_type) {
	in_skip(dwx_sizeof(value_type(value)));
	return;
    }
    s = value_string(value);
    n = strlen(s)+1;
    p = in_read(n);
    if (strcmp(s,p)) {
	fprintf(stderr,"read mismatch\n");
	exit(1);
    }
    in_free(p);
}
#endif


void stmt_if(const struct stmt *self)
{
    ip = int_eval(self->u.if_stmt.expr) ?
      self->u.if_stmt.true : self->u.if_stmt.false;
}


void stmt_break(const struct stmt *self)
{
    ip = self->u.break_stmt.loop;
}


void stmt_exit(const struct stmt *self)
{
    int status;

    if (!in_eof()) fprintf(stderr,"warning: unread dump data\n");
    status = self->u.exit.expr ? int_eval(self->u.exit.expr) : 0;
    var_cleanup();
    stmt_cleanup();
    io_cleanup();
    location_cleanup();
    report_memory_leaks();
    exit(status);
}


void stmt_die(const struct stmt *self)
{
    struct value *value;

    value = self->u.die.expr ? evaluate(self->u.die.expr) : NULL;
    if (value) fprintf(stderr,"%s\n",value_string(value));
    else clerrorf("died");
    exit(1);
}


void stmt_while(const struct stmt *self)
{
    if (int_eval(self->u.while_stmt.expr)) ip = self->u.while_stmt.body;
}


/* ----- Script execution -------------------------------------------------- */


void execute(struct stmt *pgm)
{
    ip = pgm;
    while (ip) {
	const struct stmt *stmt = ip;

//fprintf(stderr,"%p (%p)\n",ip,ip->fn);
	ip = ip->next;
	set_location(&stmt->loc);
	stmt->fn(stmt);
    }
}
